/********************************************************
Workspace: IAR_AVR
Chip type: MEGA8A
Project  : Mi_kettle
Date     : Jun, 2023
Version  : 0
*********************************************************/

#include <ioavr.h>
#include <intrinsics.h>
#include <stdbool.h>
#include <stdlib.h>

#include <math.h>

#include <bitset.c>
#include "defines.c"

#include "periph_init.c"
#include "beep.c"
#include "adc_routines.c"

#include "interrupts.c"

void devcon(u16*);

void main()
{
 cfg.gpio();
 cfg.acmp();
 cfg.t0();
 cfg.adc();

 cfg.iwdg();

 __enable_interrupt();

 ENDLESS
 {
  static u16 beep_tim;

  __watchdog_reset();

  if(beep_tim)
  {
   TCCR0 = 1<<CS01; //clk prescaler 8

   TCNT0 = 0xff - (SYSTEM_FRQ/1000/8); //1000Hz, psc_8 ->130

   beep_tim--;

   beep_processing(beep_tim);
  }
  else
  {
   TCCR0 = 1<<CS01 | 1<<CS00; //clk prescaler 64

   TCNT0 = 0xff - (SYSTEM_FRQ/100/64); //100Hz, psc_8 ->99

   for(u8 i=0; i<TIM; i++) if(Tim[i])Tim[i]--;

   devcon(&beep_tim);

   cfg.adc(true);

   MCUCR = 1<<SE | 1<<SM0; //noise reduction mode

   __sleep(); //  

   cfg.adc(false);

   MCUCR = 1<<SE; //idle mode
  }

   __sleep(); //wakeup from t0_overflow
 }
}


void devcon(u16* beep_tim)
{
 static bool itr=0;
 static enum {start, work, wait, error} dev_st_tmp,  dev_st = start;
 static enum {
              no_error,
              t_grow_error,
              adc_error,
              timeout_error,
             } err = no_error;

 if(err != no_error) dev_st = error;
 if(dev_st != dev_st_tmp)
 {
  dev_st_tmp = dev_st;

  itr = 0;
 }

 if(dev_st != work) REL_OFF;

 switch (dev_st)
 {
  //           
  case start:
  {
   static u8 stage;

   if(!itr)
   {
    itr = 1;

    stage = 0;
    Tim[SYS] = 0;
   }
   else
   {
    if(!Tim[SYS])
    {
     if(!stage)
     {
      LED_RED(1);

      Tim[SYS] = TSEC/3;
     }
     else if(stage ==1)
     {
      LED_OFF;

      Tim[SYS] = TSEC/5;
     }
     else
     {
      LED_GRN(1);

      dev_st = work;
     }

     stage++;
    }
   }
  }
  break;

  case work:
  {
   static bool boiling_complete;
   static u8 init_cnt; //      

   if(!itr)
   {
    itr = 1;
    init_cnt = 0;

    boiling_complete = false;

    Tim[SYS] = MAX_TIME;

    LED_RED(1);

    REL_ON;

    beep(beep_tim);
   }

   //       
   if(boiling_complete == true)
   {
    REL_OFF;

    if(three_beep(beep_tim, &Tim[BEEP])==true)
    {
     dev_st = wait;
    }
   }
   //  
   //      (         
   //            
   else if(!Tim[MSR])
   {
    static u8  adc_cnt = 0; //    
    static u16 adc_tmp = 0;
           u16 adc_val;

    adc_val  = ADCL;
    adc_val += ADCH<<8;

    adc_cnt++;
    adc_tmp += adc_val;

    #define AVGD 2

    //    
    if(adc_cnt >= 1<<AVGD)
    {
     #define ADC_ML 10
     static u16 adc_tcps[ADC_ML]; //   ADC_ML 
     u8 cnt_gmx = 0;              //      

     Tim[MSR] = TSEC;             //       

     //
     adc_val = adc_tmp >>AVGD;
     adc_cnt = 0;
     adc_tmp = 0;

     //   
     for(u8 i=0; i<ADC_ML-1; i++) adc_tcps[i] = adc_tcps[i+1];
     adc_tcps[ADC_ML-1] = adc_val;


     //        
     //       
     {
      //           (2)
      static u8 mtv = get_min_adc_time();
      u8 T;
      u32 r2_tmp;

      if(mtv >= ADC_ML-1) mtv = ADC_ML-1;
      if(init_cnt < 0xff) init_cnt++;

      r2_tmp = get_R2(adc_val);
      T = get_T(r2_tmp);

      //     
      for(u8 i=ADC_ML-1; i; i--)
      {
       if(adc_tcps[i] > adc_tcps[i-1] + ADC_GMX) cnt_gmx++;
      }


      #define BPV 3
      //          
      static const u8 bmtv = BPV * mtv;

      //     
      if(init_cnt < SNS_DLY + bmtv)
      {
       static u8 cnt = 0;

       if(T > 98)
       {
        if(cnt > BPV) boiling_complete = true;
        else cnt++;
       }
       else if(cnt) cnt--;
      }
      //       
      else
      {
       u8 cnt = 0;

       //     
       for(u8 i=ADC_ML-1; i>=mtv; i-=mtv)
       {
        if(adc_tcps[i] <= adc_tcps[i-mtv]) cnt++;
        else                               break;
       }

       //        
       if(cnt > BPV || T > 95) boiling_complete = true;
      }
     }

     // 
     {
      //    
      if(Tim[SYS] == 0) err = timeout_error;

      //   
      {
       static u8 avh_cnt = 0;

       if(adc_val > ADC_TOP || adc_val < ADC_BTM)
       {
        avh_cnt++;

        if(avh_cnt > 3) err = adc_error;
       }
       else if(avh_cnt) avh_cnt--;
      }

      //   
      if(cnt_gmx >= ADC_ML - 1) err = t_grow_error;
     }
    }
   }
  }
  break;

  case wait:
  {
   if(!itr)
   {
    itr = 1;

    LED_GRN(1);
   }
  }
  break;

  case error:
  {
   static int ind_stage;
   static bool ind_trg;

   if(!itr)
   {
    itr = 1;
    ind_stage = 0;
    ind_trg = false;

    Tim[ERR] = TSEC;
   }

   /*
     
        
        2      
   */
   if(!Tim[ERR])
   {
    //   
    if(ind_trg == false)
    {
     ind_trg = true;

     // -  
     if(ind_stage == 0)
     {
      Tim[ERR] = TSEC - TSEC/3;
      ind_stage ++;

      LED_RED(1);

      beep(beep_tim);
     }
     // -     
     else if(ind_stage < (int)err+1)
     {
      Tim[ERR] = TSEC/2;
      ind_stage ++;

      LED_GRN(1);
     }
     else
     {
      ind_stage = 0;
      ind_trg = false;

      Tim[ERR] = TSEC*2;
     }
    }
    //    
    else
    {
     ind_trg = false;
     Tim[ERR] = TSEC/2;

     LED_RED(0);
     LED_GRN(0);
    }
   }

   if(dev_st != error) err = no_error;
  }
  break;
 }
}

